#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "vector.h"
#include "pattern.h"

static vector_t updir   = { 0.0, 0.0,  1.0 };
static vector_t downdir = { 0.0, 0.0, -1.0 };
#define DELTA 0.0001

static int clear_points(int n, vector_t *points)
{
  int i;
  for (i = 0; i < n; i++)
    vector_zero(&points[i]);
  return 0;
}

static int clear_angles(int n, vector_t *angles)
{
  int i;
  for (i = 0; i < n; i++)
    vector_zero(&angles[i]);
  return 0;
}

static int clear(int n, vector_t *points, vector_t *angles)
{
  clear_points(n, points);
  clear_angles(n, angles);
  return 0;
}

static int pattern_default(int n, int param, vector_t *points, vector_t *angles)
{
  int i;
  clear(n, points, angles);
  for (i = 0; i < n; i++) {
    points[i].x = cos(2.0 * M_PI * i * 0.15 * param + DELTA) * 0.8;
    points[i].y = sin(2.0 * M_PI * i * 0.15 * param + DELTA) * 0.8;
    points[i].z = (double)(i + 1) / n * 2.0 - 1.0;
  }
  return 0;
}

static int pattern_random(int n, int param, vector_t *points, vector_t *angles)
{
  int i;
  clear(n, points, angles);
  for (i = 0; i < n; i++) {
    points[i].x = ((rand() % 201) - 100) * 0.01;
    points[i].y = ((rand() % 201) - 100) * 0.01;
    points[i].z = ((rand() % 201) - 100) * 0.01;
  }
  return 0;
}

static int pattern_random_angle(int n, int param, vector_t *points, vector_t *angles)
{
  int i;
  clear(n, points, angles);
  for (i = 0; i < n; i++) {
    points[i].x = ((rand() % 201) - 100) * 0.01;
    points[i].y = ((rand() % 201) - 100) * 0.01;
    points[i].z = ((rand() % 201) - 100) * 0.01;
    vector_set(&angles[i], rand() % 360, rand() % 360, rand() % 360);
  }
  return 0;
}

static int pattern_circle(int n, int param, vector_t *points, vector_t *angles)
{
  int i;
  clear(n, points, angles);
  for (i = 0; i < n; i++) {
    points[i].x = cos(2.0 * M_PI * ((double)i / n) + DELTA) * 1.5;
    points[i].y = sin(2.0 * M_PI * ((double)i / n) + DELTA) * 1.5;
  }
  return 0;
}

static int pattern_circle_center(int n, int param, vector_t *points, vector_t *angles)
{
  int i;
  vector_t dir;
  clear(n, points, angles);
  for (i = 0; i < n; i++) {
    points[i].x = cos(2.0 * M_PI * ((double)i / n) + DELTA) * 1.5;
    points[i].y = sin(2.0 * M_PI * ((double)i / n) + DELTA) * 1.5;
    vector_mul(&dir, &points[i], -1.0);
    vector_calc_angle(&angles[i], &updir, &dir);
  }
  return 0;
}

static int pattern_circle_alt(int n, int param, vector_t *points, vector_t *angles)
{
  int i;
  vector_t dir;
  clear(n, points, angles);
  for (i = 0; i < n; i++) {
    points[i].x = cos(2.0 * M_PI * ((double)i / n) + DELTA) * 1.5;
    points[i].y = sin(2.0 * M_PI * ((double)i / n) + DELTA) * 1.5;
    if (param > 1) {
      points[i].z = ((i % param) * (1.0 / (param - 1)) - 0.5) * 0.2;
    }
    vector_mul(&dir, &points[i], -1.0);
    vector_calc_angle(&angles[i], &updir, &dir);
  }
  return 0;
}

static int pattern_circle_alt_rev(int n, int param, vector_t *points, vector_t *angles)
{
  int i;
  vector_t dir;
  clear(n, points, angles);
  for (i = 0; i < n; i++) {
    points[i].x = cos(2.0 * M_PI * ((double)i / n) + DELTA) * 1.5;
    points[i].y = sin(2.0 * M_PI * ((double)i / n) + DELTA) * 1.5;
    if (param > 1) {
      points[i].z = ((i % param) * (1.0 / (param - 1)) - 0.5) * 0.2;
    }
    vector_mul(&dir, &points[i], -1.0);
    vector_calc_angle(&angles[i], (points[i].z > -DELTA) ? &updir : &downdir, &dir);
  }
  return 0;
}

static int pattern_circle_alt_out(int n, int param, vector_t *points, vector_t *angles)
{
  int i;
  clear(n, points, angles);
  for (i = 0; i < n; i++) {
    points[i].x = cos(2.0 * M_PI * ((double)i / n) + DELTA) * 1.5;
    points[i].y = sin(2.0 * M_PI * ((double)i / n) + DELTA) * 1.5;
    points[i].z = 0.0;
    vector_calc_angle(&angles[i], &points[i], &updir);
    if (param > 1) {
      points[i].z = ((i % param) * (1.0 / (param - 1)) - 0.5) * 0.7;
    }
  }
  return 0;
}

static int pattern_circle_alt_out2(int n, int param, vector_t *points, vector_t *angles)
{
  int i;
  clear(n, points, angles);
  for (i = 0; i < n; i++) {
    points[i].x = cos(2.0 * M_PI * ((double)i / n) + DELTA) * 1.5;
    points[i].y = sin(2.0 * M_PI * ((double)i / n) + DELTA) * 1.5;
    if (param > 1) {
      points[i].z = ((i % param) * (1.0 / (param - 1)) - 0.5) * 0.7;
    }
    vector_calc_angle(&angles[i], &points[i], &updir);
  }
  return 0;
}

static int pattern_screw(int n, int param, vector_t *points, vector_t *angles)
{
  int i;
  vector_t dir;
  clear(n, points, angles);
  for (i = 0; i < n; i++) {
    points[i].x = cos(M_PI * ((double)i / n) * param + DELTA) * 1.2;
    points[i].y = sin(M_PI * ((double)i / n) * param + DELTA) * 1.2;
    points[i].z = (double)i / n * 2.0 - 1.0;
    vector_mul(&dir, &points[i], -1.0);
    vector_calc_angle(&angles[i], &updir, &dir);
  }
  return 0;
}

static int pattern_screw_many(int n, int param, vector_t *points, vector_t *angles)
{
  int i;
  vector_t dir;
  clear(n, points, angles);
  for (i = 0; i < n; i++) {
    points[i].x = cos(2.0 * M_PI * ((double)(i * 2.5) / n) * param + DELTA) * 1.2;
    points[i].y = sin(2.0 * M_PI * ((double)(i * 2.5) / n) * param + DELTA) * 1.2;
    points[i].z = (double)i / n * 2.0 - 1.0;
    vector_mul(&dir, &points[i], -1.0);
    vector_calc_angle(&angles[i], &updir, &dir);
  }
  return 0;
}

static int pattern_screw_many_out(int n, int param, vector_t *points, vector_t *angles)
{
  int i;
  clear(n, points, angles);
  for (i = 0; i < n; i++) {
    points[i].x = cos(2.0 * M_PI * ((double)(i * 2.5) / n) * param + DELTA) * 1.2;
    points[i].y = sin(2.0 * M_PI * ((double)(i * 2.5) / n) * param + DELTA) * 1.2;
    points[i].z = (double)i / n * 2.0 - 1.0;
    vector_calc_angle(&angles[i], &points[i], &updir);
  }
  return 0;
}

static int pattern_2circle(int n, int param, vector_t *points, vector_t *angles)
{
  int i, num;
  double a;
  vector_t dir;
  clear(n, points, angles);
  num = ((n - 1) / param) + 1;
  for (i = 0; i < n; i++) {
    a = 2.0 * M_PI * ((double)(i % num) / num) + DELTA;
    points[i].x = cos(a) * 1.2;
    points[i].y = sin(a) * 1.2;
    points[i].z = ((i / num) - ((param - 1) * 0.5)) * 0.4;
    vector_mul(&dir, &points[i], -1.0);
    vector_calc_angle(&angles[i], &updir, &dir);
  }
  return 0;
}

static int pattern_2circle_alt(int n, int param, vector_t *points, vector_t *angles)
{
  int i, num;
  double a;
  vector_t dir;
  clear(n, points, angles);
  num = ((n - 1) / param) + 1;
  for (i = 0; i < n; i++) {
    a = 2.0 * M_PI * ((double)(i % num) / num) + DELTA;
    a += 2.0 * M_PI * (1.0 / num) * 0.5 * ((i / num) % 2);
    points[i].x = cos(a) * 1.2;
    points[i].y = sin(a) * 1.2;
    points[i].z = ((i / num) - ((param - 1) * 0.5)) * 0.4;
    vector_mul(&dir, &points[i], -1.0);
    vector_calc_angle(&angles[i], &updir, &dir);
  }
  return 0;
}

static int pattern_2circle_alt_out(int n, int param, vector_t *points, vector_t *angles)
{
  int i, num;
  double a;
  clear(n, points, angles);
  num = ((n - 1) / param) + 1;
  for (i = 0; i < n; i++) {
    a = 2.0 * M_PI * ((double)(i % num) / num) + DELTA;
    a += 2.0 * M_PI * (1.0 / num) * 0.5 * ((i / num) % 2);
    points[i].x = cos(a) * 1.2;
    points[i].y = sin(a) * 1.2;
    points[i].z = ((i / num) - ((param - 1) * 0.5)) * 0.4;
    vector_calc_angle(&angles[i], &points[i], (points[i].z > -DELTA) ? &updir : &downdir);
  }
  return 0;
}

static int pattern_bridge(int n, int param, vector_t *points, vector_t *angles)
{
  int i, num;
  double a;
  vector_t dir;
  clear(n, points, angles);
  num = (n + (param - 1)) / param;
  for (i = 0; i < n; i++) {
    a = 2.0 * M_PI * (i / num) / param + DELTA;
    points[i].x = cos(a) * 1.2;
    points[i].y = sin(a) * 1.2;
    points[i].z = ((i % num) - (num - 1) * 0.5) * 0.5;
    vector_mul(&dir, &points[i], -1.0);
    vector_calc_angle(&angles[i], &updir, &dir);
  }
  return 0;
}

static int pattern_bridge_alt(int n, int param, vector_t *points, vector_t *angles)
{
  int i, num;
  double a;
  vector_t dir;
  clear(n, points, angles);
  num = (n + (param - 1)) / param;
  for (i = 0; i < n; i++) {
    a = 2.0 * M_PI * (i / num) / param + DELTA;
    points[i].x = cos(a) * 1.2;
    points[i].y = sin(a) * 1.2;
    points[i].z = ((i % num) - (num - 1) * 0.5) * 0.5;
    points[i].z += (i / num) * 0.2;
    vector_mul(&dir, &points[i], -1.0);
    vector_calc_angle(&angles[i], &updir, &dir);
  }
  return 0;
}

static int pattern_bridge_alt_out(int n, int param, vector_t *points, vector_t *angles)
{
  int i, num;
  double a;
  clear(n, points, angles);
  num = (n + (param - 1)) / param;
  for (i = 0; i < n; i++) {
    a = 2.0 * M_PI * (i / num) / param + DELTA;
    points[i].x = cos(a) * 1.2;
    points[i].y = sin(a) * 1.2;
    points[i].z = 0.0;
    vector_calc_angle(&angles[i], &points[i], &updir);
    points[i].z = ((i % num) - (num - 1) * 0.5) * 0.5;
    points[i].z += (i / num) * 0.2;
  }
  return 0;
}

static int pattern_bridge_alt_out2(int n, int param, vector_t *points, vector_t *angles)
{
  int i, num;
  double a;
  clear(n, points, angles);
  num = (n + (param - 1)) / param;
  for (i = 0; i < n; i++) {
    a = 2.0 * M_PI * (i / num) / param + DELTA;
    points[i].x = cos(a) * 1.2;
    points[i].y = sin(a) * 1.2;
    points[i].z = ((i % num) - (num - 1) * 0.5) * 0.5;
    points[i].z += (i / num) * 0.2;
    vector_calc_angle(&angles[i], &points[i], &updir);
  }
  return 0;
}

static int pattern_tower(int n, int param, vector_t *points, vector_t *angles)
{
  int i;
  clear(n, points, angles);
  for (i = 0; i < n; i++) {
    points[i].z = (double)i / n * 2.0 - 1.0;
  }
  return 0;
}

static int pattern_flat(int n, int param, vector_t *points, vector_t *angles)
{
  int i, r = 0, p = 0, num = 0;
  clear(n, points, angles);
  for (i = 0; i < n; i++) {
    if (p == 0) {
      num = (int)((2.0 * M_PI) * r);
      if (num == 0) num = 1;
      if (num > n - i)
	num = n - i;
    }
    points[i].x = cos(2.0 * M_PI * ((double)p / num)) * r;
    points[i].y = sin(2.0 * M_PI * ((double)p / num)) * r;
    if (r == 0)
      points[i].x += DELTA;
    p++;
    if (p >= num) {
      r++;
      p = 0;
    }
  }
  return 0;
}

static int pattern_flat_center(int n, int param, vector_t *points, vector_t *angles)
{
  int i;
  vector_t dir;
  pattern_flat(n, param, points, angles);
  for (i = 0; i < n; i++) {
    vector_mul(&dir, &points[i], -1.0);
    vector_calc_angle(&angles[i], &updir, &dir);
  }
  return 0;
}

static int pattern_flat_cone(int n, int param, vector_t *points, vector_t *angles)
{
  int i;
  double a;
  vector_t dir;
  pattern_flat(n, param, points, angles);
  for (i = 0; i < n; i++) {
    a = sqrt(points[i].x * points[i].x + points[i].y * points[i].y);
    points[i].z = 1.0 - a * 0.2;
    vector_mul(&dir, &points[i], -1.0);
    vector_calc_angle(&angles[i], &updir, &dir);
  }
  return 0;
}

static int pattern_flat_cone_out(int n, int param, vector_t *points, vector_t *angles)
{
  int i;
  double a;
  vector_t dir;
  pattern_flat(n, param, points, angles);
  for (i = 0; i < n; i++) {
    a = sqrt(points[i].x * points[i].x + points[i].y * points[i].y);
    vector_mul(&dir, &points[i], -1.0);
    vector_calc_angle(&angles[i], &points[i], &updir);
    points[i].z = 1.0 - a * 0.2;
  }
  return 0;
}

static int pattern_flat_mount(int n, int param, vector_t *points, vector_t *angles)
{
  int i;
  double a;
  vector_t dir;
  pattern_flat(n, param, points, angles);
  for (i = 0; i < n; i++) {
    a = sqrt(points[i].x * points[i].x + points[i].y * points[i].y);
    points[i].z = 1.0 / (1 + a * a);
    vector_mul(&dir, &points[i], -1.0);
    vector_calc_angle(&angles[i], &updir, &dir);
  }
  return 0;
}

static int pattern_flat_mount_slope(int n, int param, vector_t *points, vector_t *angles)
{
  int i;
  double a;
  vector_t dir, norm;
  pattern_flat(n, param, points, angles);
  for (i = 0; i < n; i++) {
    a = sqrt(points[i].x * points[i].x + points[i].y * points[i].y);
    points[i].z = 1.0 / (1 + a * a);
    vector_copy(&norm, &points[i]);
    if (i > 0) {
      a += DELTA;
      a = 1.0 / (1 + a * a);
      a = DELTA / (points[i].z - a);
      norm.z = sqrt(norm.x * norm.x + norm.y * norm.y) * a;
    }
    vector_mul(&dir, &points[i], -1.0);
    vector_calc_angle(&angles[i], &norm, &dir);
  }
  return 0;
}

static int pattern_square(int n, int param, vector_t *points, vector_t *angles)
{
  int i;
  clear(n, points, angles);
  for (i = 0; i < n; i++) {
    points[i].x = ((i % param) - (param - 1) * 0.5) * 0.8;
    points[i].z = ((i / param) - (((n + param - 1) / param) - 1) * 0.5) * 0.4;
  }
  return 0;
}

static int pattern_square_alt(int n, int param, vector_t *points, vector_t *angles)
{
  int i;
  clear(n, points, angles);
  for (i = 0; i < n; i++) {
    points[i].x = ((i % param) - (param - 1) * 0.5) * 0.8;
    points[i].z = ((i / param) - (((n + param - 1) / param) - 1) * 0.5) * 0.4;
    points[i].z += (i % 2) * 0.1;
  }
  return 0;
}

static int pattern_circle_virt(int n, int param, vector_t *points, vector_t *angles)
{
  int i;
  clear(n, points, angles);
  for (i = 0; i < n; i++) {
    points[i].x = cos(2.0 * M_PI * ((double)i / n) + DELTA) * 1.2;
    points[i].z = sin(2.0 * M_PI * ((double)i / n) + DELTA) * 1.2;
  }
  return 0;
}

static int pattern_circle_virt_double(int n, int param, vector_t *points, vector_t *angles)
{
  int i, num;
  double a;
  clear(n, points, angles);
  points[0].z = -1.2;
  if (n > 0)
    points[n - 1].z = 1.2;
  if (n > 1) {
    num = ((n - 2) + param - 1) / param;
    for (i = 1; i < n - 1; i++) {
      a = M_PI * (1.0 / (num + 1)) * ((i - 1 + param) / param) + DELTA;
      a -= M_PI * 0.5;
      points[i].x = cos(a) * 1.2;
      points[i].z = sin(a) * 1.2;
      a = 360.0 * ((i - 1) % param) / param;
      vector_rotate_z(&points[i], &points[i], a);
    }
  }
  return 0;
}

static int pattern_circle_virt_double_center(int n, int param, vector_t *points, vector_t *angles)
{
  int i;
  vector_t dir;
  pattern_circle_virt_double(n, param, points, angles);
  for (i = 0; i < n; i++) {
    vector_mul(&dir, &points[i], -1.0);
    vector_calc_angle(&angles[i], &updir, &dir);
  }
  return 0;
}

static int pattern_circle_virt_double_out(int n, int param, vector_t *points, vector_t *angles)
{
  int i;
  pattern_circle_virt_double(n, param, points, angles);
  for (i = 0; i < n; i++) {
    vector_calc_angle(&angles[i], &points[i], &updir);
  }
  return 0;
}

static int pattern_ball_random(int n, int param, vector_t *points, vector_t *angles)
{
  int i, j, num;
  vector_t vd, v;
  double r;

  vector_set(&points[0], 0.0, 0.0, 1.0);
  clear_angles(n, angles);

  num = 128;

  while (1) {
    for (i = 1; i < n; i++) {
      vector_zero(&v);
      for (j = 0; j < n; j++) {
	if (i == j)
	  continue;
	vector_sub(&vd, &points[i], &points[j]);
	while ((r = vector_length(&vd)) == 0.0)
	  vector_rand(&vd);
	vector_normalize(&vd, 1.0 / (r*r));
	vector_add(&vd, &vd, &points[i]);
	vector_normalize(&vd, 1.0);
	vector_sub(&vd, &vd, &points[i]);
	vector_add(&v, &v, &vd);
      }
      if (vector_length(&v) > 0.5)
	vector_normalize(&v, 0.5);
      vector_add(&points[i], &points[i], &v);
      vector_normalize(&points[i], 1.0);
      num--;
    }
    if (num < 0)
      break;
  }

  return 0;
}

static int pattern_ball_random_center(int n, int param, vector_t *points, vector_t *angles)
{
  int i;
  vector_t dir;
  pattern_ball_random(n, param, points, angles);
  for (i = 0; i < n; i++) {
    vector_mul(&dir, &points[i], -1.0);
    vector_calc_angle(&angles[i], &updir, &dir);
  }
  return 0;
}

static int pattern_ball_random_out(int n, int param, vector_t *points, vector_t *angles)
{
  int i;
  pattern_ball_random(n, param, points, angles);
  for (i = 0; i < n; i++)
    vector_calc_angle(&angles[i], &points[i], &updir);
  return 0;
}

typedef int (*pattern_func_t)(int n, int param, vector_t *points, vector_t *angles);
static pattern_func_t funcs[] = {
  pattern_default,
  pattern_random, pattern_random_angle,
  pattern_circle, pattern_circle_center,
  pattern_circle_alt, pattern_circle_alt_rev,
  pattern_circle_alt_out, pattern_circle_alt_out2,
  pattern_screw, pattern_screw_many, pattern_screw_many_out,
  pattern_2circle, pattern_2circle_alt, pattern_2circle_alt_out,
  pattern_bridge, pattern_bridge_alt,
  pattern_bridge_alt_out, pattern_bridge_alt_out2,
  pattern_tower,
  pattern_flat, pattern_flat_center, pattern_flat_cone, pattern_flat_cone_out,
  pattern_flat_mount, pattern_flat_mount_slope,
  pattern_square, pattern_square_alt,
  pattern_circle_virt, pattern_circle_virt_double,
  pattern_circle_virt_double_center,
  pattern_circle_virt_double_out,
  pattern_ball_random, pattern_ball_random_center, pattern_ball_random_out,
};

int pattern_set(int pattern, int n, int param, vector_t *points, vector_t *angles)
{
  pattern_func_t f;
  int num;

  num = sizeof(funcs) / sizeof(funcs[0]);
  while (pattern < 0) pattern += num;
  f = funcs[pattern % num];
  f(n, param, points, angles);

  return 0;
}
